#! /usr/bin/python

from __future__ import division
import os, time
import pylab
import itertools
import scipy.misc
import numpy as np
from ctypes import *
from NN import NN

def timer(f):
   def wrap(*args):
      time1 = time.time()
      ret = f(*args)
      time2 = time.time()
      print("{} took {:.3f} ms".format(f.func_name, (time2-time1)*1000.0))
      return ret
   return wrap

class denoise:
   # athreshold: combined with sigma for affine region selection: 1
   # start: first similar group size for affine region selection: 1
   def __init__(self, cleandir, sigma, athreshold,
                      pSize, wSize, start, EMruns,
                      nFriends, imdim, gamma, alpha, tol):
      # algo param section
      self.sigma = sigma # input noise
      self.pSize = pSize # patch size
      self.wSize = wSize # search window size
      self.EMruns = EMruns # GMM EM rounds
      self.start = start
      self.athreshold = athreshold
      self.nFriends = nFriends # minimial friends needed for structure
      self.imdim = imdim # implicit dimension in autosimilar denoising
      self.gamma = gamma
      self.alpha = alpha
      self.tol = tol
      self.test = True  # operating mode
      # margin for interpreting patch classification pixel-wise
      # although pSize can be either odd or even, let's make
      # left up margin smaller than right down margin
      self.lumargin = int((pSize-1)/2)
      self.rdmargin = int((pSize-1+((pSize+1)%2))/2)
      self.dSize = self.pSize**2
      # SPLE GMM convention
      self.model_texture = 18
      self.model_affine = 19
      # get clean, noisy and extended noisy images
      self.im_paths = [os.path.join(cleandir, f)
                       for f in os.listdir(cleandir) if f.endswith(".png")]
      self.clean_ims = [scipy.misc.imread(p) for p in self.im_paths]
      self.getNoisy()
      # pmap may get modified in the process
      self.getPmaps()

   def getNoisy(self):
      self.noisy_ims = []
      self.ext_noisy_ims = []
      for clean in self.clean_ims:
         np.random.seed(100)
         noisy = clean+np.random.normal(scale=self.sigma, size=clean.shape)
         self.noisy_ims.append(noisy)
         self.ext_noisy_ims.append(self.extendIm(noisy))

   def getPmaps(self):
      self.pmaps = []
      for index in range(len(self.ext_noisy_ims)):
         self.pmaps.append(self.classifyPixels(index))

   def classifyPixels(self, index):
      func = CDLL("./classify.so").classify
      # the func interface
      c_double_p = POINTER(c_double)
      func.argtypes = [c_double_p, c_int, c_int, c_double, c_int, c_int]
      clean = self.clean_ims[index]
      func.restype = np.ctypeslib.ndpointer(dtype=c_int, shape=clean.shape)
      # compute its patch map
      ext_noisy = self.ext_noisy_ims[index]
      ext_rows, ext_cols = ext_noisy.shape
      pmap = func(ext_noisy.ctypes.data_as(c_double_p), ext_rows, ext_cols,
                  self.sigma, self.pSize, self.EMruns)
      if self.test:
         affine = pmap == self.model_affine
         texture = pmap == self.model_texture
         structure = pmap < self.model_texture
         denoise.display([clean, affine, texture, structure],
                         ["clean", "affine", "texture", "structure"],
                         ["gray", "gray", "gray", "gray"], samePanel=True)
         scipy.misc.imsave("affinep.png", affine*250)
      return pmap

   @timer
   def denoiseImage(self, index):
      weighted, denoised, reassign = self.denoiseAffine(index)
      # weighteds, denoiseds, reassign = self.denoiseStruct(index)
      # weightedt, denoisedt = self.denoiseTexture(index)
      # denoised = (denoiseda + denoiseds + denoisedt)/(weighteda + weighteds + weightedt)
      if self.test:
         assert(np.mean(self.pmaps[index][reassign==1])==self.model_texture)
         # when testing individual denoising algo, pick one criterion below
         # though the result will always be called affine
         affine = self.pmaps[index] == self.model_affine  # affine
         # affine = self.pmaps[index] < self.model_texture  # oriented
         # affine = self.pmaps[index] == self.model_texture  # texture
         a_denoised = denoised/np.maximum(weighted, np.ones(weighted.shape))
         diff = (self.clean_ims[index]-a_denoised)*affine
         rmse = np.sqrt((diff**2).sum()/affine.sum())
         print("rmse from affine regression is {0:.2f}".format(rmse))
         # however, texture doesn't have a reassign
         denoise.display([self.clean_ims[index], reassign, a_denoised],
                         ["clean", "affine reassigned to texture", "affine rmse = {0:.2f} vs. input sigma {1:.2f}".format(rmse, self.sigma)],
                         ["gray", "gray", "gray"], samePanel=True)
         scipy.misc.imsave("affineReassign.png", reassign)
         scipy.misc.imsave("affineONLY.png", a_denoised)
         scipy.misc.imsave("affinenp.png", affine*250)

   def denoiseTexture(self, index):
      im_rows, im_cols = self.pmaps[index].shape
      pset = [p for p in itertools.product(range(im_rows), range(im_cols)) if self.pmaps[index][p] == self.model_texture]
      patches = self.im2patches(self.ext_noisy_ims[index])
      noisy = np.asarray([patches[p] for p in pset]).T
      # neural denoising
      net = "GPU_NN"
      nn = NN(net, self.sigma)
      estimates = nn.feedforward(noisy, self.sigma, areClean=False)
      # fill in
      weighted = np.zeros(self.ext_noisy_ims[index].shape)
      denoised = np.zeros(self.ext_noisy_ims[index].shape)
      for k, p in enumerate(pset):
         r, c = p
         weighted[r:r+self.pSize, c:c+self.pSize] += 1
         denoised[r:r+self.pSize, c:c+self.pSize] += estimates[:, k].reshape(self.pSize, self.pSize)

      return weighted[self.lumargin:self.lumargin+im_rows, self.lumargin:self.lumargin+im_cols], denoised[self.lumargin:self.lumargin+im_rows, self.lumargin:self.lumargin+im_cols]

   def denoiseStruct(self, index):
      im_rows, im_cols = self.pmaps[index].shape
      reassign = np.zeros([im_rows, im_cols])
      # reframe the image at the end
      weighted = np.zeros(self.ext_noisy_ims[index].shape)
      denoised = np.zeros(self.ext_noisy_ims[index].shape)
      patches = self.im2patches(self.ext_noisy_ims[index])
      # two stages: Enet, combined with some heuristics for friend selection
      # a patch might be brought to texture in the process for lack of friends
      # once confirmed, a filter then acts on the selected group to yield an estimate
      for mypos in itertools.product(range(im_rows), range(im_cols)):
         if self.pmaps[index][mypos] < self.model_texture:
            group = denoise.sortNonLocal(mypos, patches, self.wSize, actualSort=False)
            # retain the oriented patches ONLY for a speed-up
            sgroup = [guy for guy in group[1:] if self.pmaps[index][guy["pos"]] < self.model_texture]
            # it's possible to have sgroup empty here too
            if len(sgroup) < self.nFriends:
               reassign[mypos] = 1
               self.pmaps[index][mypos] = self.model_texture
               continue
            # further shrink the group to those POSITIVELY correlated with myself
            y = patches[group[0]["pos"]].reshape(self.dSize, 1)
            X = np.asarray([patches[guy["pos"]] for guy in sgroup]).T
            corr = np.dot((y - np.mean(y)).T, X - np.tile(np.mean(X, axis=0), [self.dSize, 1])).flatten()
            sgroup = [sgroup[k] for k in range(len(corr)) if corr[k] > 0]
            # what if there aren't enough friends
            if len(sgroup) < self.nFriends:
               reassign[mypos] = 1
               self.pmaps[index][mypos] = self.model_texture
               continue
            # elastic net variable selection by shooting
            # based on patch-wide variation not their means
            X = np.asarray([X[:, k] for k in range(len(corr)) if corr[k] > 0]).T
            sol = denoise.shootingEnet(y, X, self.gamma, self.alpha, tol=self.tol).flatten()
            select = [sgroup[k] for k in range(len(sol)) if sol[k] != 0]
            if len(select) < self.nFriends:
               reassign[mypos] = 1
               self.pmaps[index][mypos] = self.model_texture
               continue
            select.insert(0, group[0])
            # now act on the selected group
            estimate = self.runFilter(select, patches)
            # remove False to see rolling examples
            if self.test and False:
               print("reference patch {} belonging to model {}".format(group[0]["pos"], self.pmaps[index][group[0]["pos"]]))
               print("from the initial group of {} oriented and positive correlated members,".format(len(sgroup)))
               print("I selected {} friends belonging to model {}".format(len(select)-1, [self.pmaps[index][m["pos"]] for m in select[1:]]))
               # more control over the test display
               row, col = group[0]["pos"]
               if row > self.pSize and col > self.pSize:
                  self.highlight(index, select)
                  NN.showAtoms(np.asarray([patches[select[0]["pos"]], estimate]).T, "before vs. after", enhance=False)
               # this continue parades the oriented autosimilar group faster
               continue
            # fill in
            r, c = mypos
            weighted[r:r+self.pSize, c:c+self.pSize] += 1
            denoised[r:r+self.pSize, c:c+self.pSize] += estimate.reshape(self.pSize, self.pSize)

      return weighted[self.lumargin:self.lumargin+im_rows, self.lumargin:self.lumargin+im_cols], denoised[self.lumargin:self.lumargin+im_rows, self.lumargin:self.lumargin+im_cols], reassign

   def runFilter(self, select, patches):
      # wiener filter
      noisy = np.asarray([patches[guy["pos"]] for guy in select]).T
      nmeans = np.mean(noisy, axis=0)
      noisy = noisy - np.tile(nmeans, [self.dSize, 1])
      # implicit dimension (my belief)
      d, base = denoise.truncatedEigen(noisy, self.imdim)
      ratio = d/(d+self.sigma**2)
      wproj = np.dot(base.T, noisy[:, 0]) * ratio
      return np.dot(base, wproj.reshape(self.imdim, 1)) + nmeans[0]

   def denoiseAffine(self, index):
      im_rows, im_cols = self.pmaps[index].shape
      weighted = np.zeros([im_rows, im_cols])
      denoised = np.zeros([im_rows, im_cols])
      reassign = np.zeros([im_rows, im_cols])
      patches = self.im2patches(self.ext_noisy_ims[index])
      for mypos in itertools.product(range(im_rows), range(im_cols)):
         if self.pmaps[index][mypos] == self.model_affine and weighted[mypos]==0:
            # linear regression heuristic
            neigbhors = denoise.sortNonLocal(mypos, patches, self.wSize)
            pset = set() # retained region for mypos in the noisy image
            # track list
            lpset = [] # list of pset I've looked at
            llreg = [] # corresponding regression results
            # self.start specifies the size of the first block
            # for the remaining region expansions, increase by 1 patch at a time
            begin = 0 # end = self.start
            for end in range(self.start, len(neigbhors)):
               for guy in neigbhors[begin:end]:
                  r, c = guy["pos"]
                  # (r,c) is the coordinate in the clean image
                  # as we collect pixels, make sure they lie in the correct domain
                  mr = max(r-self.lumargin, 0)
                  mc = max(c-self.lumargin, 0)
                  xr = min(r+self.rdmargin+1, im_rows)
                  xc = min(c+self.rdmargin+1, im_cols)
                  pset.update(itertools.product(range(mr, xr), range(mc, xc)))
               # new pixel installment
               lpset.append(list(pset))
               nb_pos = len(lpset[-1])
               # add one neighbor at a time from now on
               begin = end
               # affine regression on the most recent pixel collection
               vals = np.asarray([self.noisy_ims[index][p] for p in lpset[-1]]).reshape(nb_pos, 1)
               llreg.append(denoise.lregression(vals, lpset[-1]))
               residu = llreg[-1]["residu"][0]/np.sqrt(nb_pos)
               # when to call it quits
               if residu > self.sigma * self.athreshold:
                  break
            # record the denoised region and denoised pixel values
            if(len(lpset) == 1):
               # trust affine only to an extent
               reassign[mypos] = 1
               self.pmaps[index][mypos] = self.model_texture
            else:
               for k, pos in enumerate(lpset[-2]):
                  weighted[pos] += 1
                  denoised[pos] += llreg[-2]["denoised"][k]

      return weighted, denoised, reassign

   def extendIm(self, img):
      # extend the image horizontally
      lpart = np.fliplr(img[:, 1:self.lumargin+1])
      rpart = np.fliplr(img[:, -1-self.rdmargin:-1])
      img = np.concatenate((lpart, img, rpart), axis=1)
      # extend the image vertically
      tpart = np.flipud(img[1:self.lumargin+1, :])
      bpart = np.flipud(img[-1-self.rdmargin:-1, :])
      img = np.concatenate((tpart, img, bpart), axis=0)
      return img

   @staticmethod
   def display(imgs, titles, cmaps, samePanel=True):
      nDisplay = len(imgs)
      if samePanel:
         f, axarr = pylab.subplots(nDisplay, sharex=False)
         for k in range(nDisplay):
            axarr[k].imshow(imgs[k], cmaps[k])
            axarr[k].set_title(titles[k])
            axarr[k].axis("off")
         pylab.show()
      else:
         for k in range(nDisplay):
            pylab.imshow(imgs[k], cmaps[k])
            pylab.title(titles[k])
            pylab.axis("off")
            pylab.show()

   @staticmethod
   def sortNonLocal(mypos, patches, wSize, actualSort=True):
      im_row, im_col = mypos
      mr = max(im_row-wSize, 0)
      mc = max(im_col-wSize, 0)
      xr = min(im_row+wSize, patches.shape[0])
      xc = min(im_col+wSize, patches.shape[1])
      if actualSort:
         # real non local means style sorting
         me = patches[mypos]
         neigbhors = [{"pos": p, "diff": ((patches[p]-me)**2).sum()}
                      for p in itertools.product(range(mr, xr), range(mc, xc)) if p != mypos]
         neigbhors.sort(key = lambda guy: guy["diff"])
         neigbhors.insert(0, {"pos": mypos, "diff": 0})
      else:
         # just return the coordinates with little cost
         neigbhors = [{"pos": p} for p in itertools.product(range(mr, xr), range(mc, xc)) if p != mypos]
         neigbhors.insert(0, {"pos": mypos})

      return neigbhors

   @staticmethod
   def lregression(inputs, poss, order=1):
      # inputs regressed on a polynomial model
      # poss is the coordinate grid associated with values in inputs
      # not generic as the polynomial order can only be either 1 or 2
      regressor = np.ones((inputs.shape[0], 3*order))
      # standardize
      minr = min([pos[0] for pos in poss])
      minc = min([pos[1] for pos in poss])
      poss = [(pos[0]-minr, pos[1]-minc) for pos in poss]
      if order <= 2:
         regressor[:, 1] = np.asarray([pos[0] for pos in poss]).flatten()
         regressor[:, 2] = np.asarray([pos[1] for pos in poss]).flatten()
         if order == 2:
            regressor[:, 3] = np.asarray([pos[0]**2 for pos in poss]).flatten()
            regressor[:, 4] = np.asarray([pos[1]**2 for pos in poss]).flatten()
            regressor[:, 5] = np.asarray([pos[0]*pos[1] for pos in poss]).flatten()
      else:
         raise IndexError("regression polynomial order cannot be {}".format(order))
      # square error minimization
      coefs = np.linalg.solve(np.dot(regressor.T, regressor), np.dot(regressor.T, inputs))
      denoised = np.dot(regressor, coefs)
      residu = np.linalg.norm(denoised - inputs, axis=0)
      return {"residu": residu, "denoised": denoised}

   def im2patches(self, img):
      # memory strides requires this
      img = np.ascontiguousarray(img)
      # make a four dimensional matrix out of img
      # the first two dimension indexes individual patch"s coordinates
      # the last two indexes the pixels within these patches
      rows, cols = img.shape
      shape = (rows-self.pSize+1, cols-self.pSize+1, self.pSize, self.pSize)
      # img is inherently a continuum of memory
      # tell where each pixel lies in the previouly defined structure
      # python is row-major by default
      strides = np.array([cols, 1, cols, 1])
      # the byte length of each item in img
      strides = strides*img.itemsize
      # form the four dimensional matrix
      blocks = np.lib.stride_tricks.as_strided(img, shape=shape, strides=strides)
      return blocks.reshape(shape[0], shape[1], self.dSize)

   def paint(self, img, pset, colors):
      cmarks = {"red": 0, "green": 1, "blue": 2}
      for k, pos in enumerate(pset):
         # differentiate the reference patch from the rest
         cmark = cmarks[colors[0]] if k == 0 else cmarks[colors[1]]
         r, c = pos
         for ch in range(3):
            if ch == cmark:
               img[r, c:c+self.pSize+1, ch] = 255
               img[r+self.pSize, c:c+self.pSize+1, ch] = 255
               img[r:r+self.pSize+1, c, ch] = 255
               img[r:r+self.pSize+1, c+self.pSize, ch] = 255
            else:
               img[r, c:c+self.pSize+1, ch] = 0
               img[r+self.pSize, c:c+self.pSize+1, ch] = 0
               img[r:r+self.pSize+1, c, ch] = 0
               img[r:r+self.pSize+1, c+self.pSize, ch] = 0
      return img

   def highlight(self, index, group):
      # only deal with grayscale images
      ext_rows, ext_cols = self.ext_noisy_ims[index].shape
      color = np.zeros([ext_rows, ext_cols, 3])
      for ch in range(3):
         color[:, :, ch] = self.ext_noisy_ims[index]
      # highlight
      pset = [guy["pos"] for guy in group]
      color = self.paint(color, pset, ["red", "green"])
      np.clip(color, 0, 255, out=color)
      denoise.display([color.astype(np.uint8)], ["selected group and the ref."],
                      ["gray"], samePanel=False)
      # display individual patches in the same panel
      patches = self.im2patches(self.ext_noisy_ims[index])
      individuals = np.asarray([patches[p] for p in pset]).T
      NN.showAtoms(individuals, "group as it is (the ref. comes the first)", enhance=False)
      NN.showAtoms(individuals, "enhanced group (the ref. comes the first)", enhance=True)

   @staticmethod
   def standardize(vec):
      rows, cols = vec.shape
      # -= can induce unintentional truncating if vec is an integer vec
      vec = vec - np.tile(np.mean(vec, axis=0), [rows, 1])
      vec = vec / np.tile(np.sqrt(np.sum(vec**2, axis=0)), [rows, 1])
      return vec

   @staticmethod
   def shootingEnet(y, X, gamma, alpha, tol=1e-10, standard_inputs=True):
      # a generic elastic net for variable selection
      # .5*||y-X*beta||_2^2 + .5*alpha*||beta||_2^2 + gamma*||beta||_1
      # the applied solver is called shooting (coordinate descent)
      if standard_inputs:
         y = denoise.standardize(y)
         X = denoise.standardize(X)
      rows, cols = X.shape
      # precompute stuff for data fitting evaluation
      multiplier = np.zeros([cols, cols])
      for k in range(cols):
         Z = X.copy()
         Z[:, k] = y.flatten()
         multiplier[k, :] = np.dot(X[:, k].T, Z)
      consts = np.sum(X**2, axis=0)+alpha
      # iterative procedure's stopping condition
      # tol is another
      maxIter = int(1e4)
      # over-completeness reigns
      curr_sol = np.zeros([cols, 1])
      # start coordinate descent
      for i in range(maxIter):
         sol = curr_sol.copy()
         for k in range(cols):
            sol[k] = -1
            S = np.dot(multiplier[k, :], sol)*-1
            if S-gamma > 0:
               sol[k] = (S-gamma)/consts[k]
            elif S+gamma < 0:
               sol[k] = (S+gamma)/consts[k]
            else:
               sol[k] = 0
         if np.linalg.norm(sol-curr_sol) < tol:
            break
         else:
            curr_sol = sol
      return sol

   @staticmethod
   def truncatedEigen(group, num):
      rows, cols = group.shape
      # two operating modes
      if cols < rows:
         Q, R = np.linalg.qr(group)
         F = np.dot(Q.T, group)
         C = np.dot(F, F.T)
      else:
         C = np.dot(group, group.T)
      # hopefully, on a smaller scale
      d, V = np.linalg.eigh(C)
      rnum = -1*num
      if cols < rows:
         vecs = np.dot(Q, V[:, rnum:])
      else:
         vecs = V[:, rnum:]
      return d[rnum:], vecs


if __name__ == "__main__":

   cleandir = "."
   sigma = 10
   pSize = 7
   wSize = 20
   start = 5
   EMruns = 1
   athreshold = 1
   nFriends = 5
   imdim = 1
   gamma = 0.4
   alpha = 0.01
   tol = 1e-3
   solver = denoise(cleandir, sigma, athreshold,
                    pSize, wSize, start, EMruns,
                    nFriends, imdim, gamma, alpha, tol)

   if False:
      import cProfile
      cProfile.run("solver.denoiseImage(0)")
   else:
      solver.denoiseImage(0)
